#
#   Autuconf macroses for checking for Berkeley DB
#   Written by Andrew W. Nosenko <awn@bcs.zp.ua>
#
#
#   Public available macro:
#       DS_BERKELEY_DB3([db_cppflags_out], [db_libs_out],
#                      [db_version], [db_version_major],
#                      [additional-action-if-found],
#                      [additional-action-if-not-found]
#                      )
#
#   Another macros are considered as private for implementation and
#   sould not be used in the configure.ac.  At this time these are:
#       DS_BERKELEY_DB3_HEADERS()
#       DS_BERKELEY_DB3_LIBS()
#

#   DS_BERKELEY_DB3_HEADERS([db_cppflags_out],
#                  [additional-action-if-found],
#                  [additional-action-if-not-found]
#                 )
AC_DEFUN(DS_BERKELEY_DB3_HEADERS,
[
ds_db3_headers_save_CPPFLAGS="$CPPFLAGS"
ds_db3_headers_CPPFLAGS=''
ds_db3_headers_success=yes

AC_ARG_WITH(db-includes,
            [  --with-db3-includes=DIR      Where to find Berkeley DB headers])
AC_MSG_CHECKING([where to find Berkeley DB headers])
if test x"$with_db3_includes" = x
then
    AC_MSG_RESULT([compiler default paths])
else
    AC_MSG_RESULT([$with_db3_includes])
    if ! test -d "$with_db3_includes"
    then
        AC_MSG_ERROR([required include path for db headers $with_db3_includes is not a directory])
    fi
    ds_db3_headers_CPPFLAGS="-I$with_db3_includes"
    CPPFLAGS="$ds_db3_headers_CPPFLAGS $CPPFLAGS"
    ds_db3_headers_CFLAGS="-I$with_db3_includes"
    CFLAGS="$ds_db3_headers_CFLAGS $CFLAGS"
fi
AC_CHECK_HEADER([db.h],
                [],
                [ ds_db3_headers_success=no ])
if test x"$ds_db3_headers_success" = xyes
then
    AC_PREPROC_IFELSE([AC_LANG_SOURCE([[
    #include <db.h>
    #if DB_VERSION_MAJOR == 3
    /* Success */
    #else
    #error Unsupported version of Berkeley DB (v3 is required)
    #endif
            ]])],
            [],
            [
                AC_MSG_FAILURE([Unsupported version of Berkeley DB (v3 is required - $DB_VERSION_MAJOR found)])
                ds_db3_headers_success=no
            ])
fi

CPPFLAGS="$ds_db3_headers_save_CPPFLAGS"
if test x"$ds_db3_headers_success" = xyes
then
    ifelse([$1], [], [:], [$1="$ds_db3_headers_CPPFLAGS"])
    ifelse([$2], [], [:], [$2])
else
    ifelse([$3], [], [:], [$3])
fi
])

#
#   DS_BERKELEY_DB3_LIBS([db_libs_out],
#                  [additional-action-if-found],
#                  [additional-action-if-not-found]
#                 )
AC_DEFUN(DS_BERKELEY_DB3_LIBS,
[
ds_db3_libs_save_LIBS="$LIBS"
ds_db3_libs_LIBS=''
ds_db3_libs_success=no

AC_ARG_WITH(db-libraries,
            [  --with-db3-libraries=DIR     Where to find Berkeley DB libraries])
AC_MSG_CHECKING([where to find Berkeley DB libraries])
if test x"$with_db3_libraries" = x
then
    AC_MSG_RESULT([compiler default paths])
else
    AC_MSG_RESULT([$with_db3_libraries])
    if ! test -d "$with_db3_libraries"
    then
        AC_MSG_ERROR([required path for db libraries ($with_db3_includes) is not a directory])
    fi
    ds_db3_libs_LIBS="-L$with_db3_libraries"
fi

AC_MSG_CHECKING([for db_create in -ldb3])
LIBS="$ds_db3_libs_LIBS -ldb3 $ds_db3_libs_save_LIBS"
AC_LINK_IFELSE([AC_LANG_PROGRAM([[
#include <db.h>
    ]],
    [[
        DB** dummy_dbp = NULL;
        db_create(dummy_dbp, NULL, 0);
    ]])],
    [
        ds_db3_libs_success=yes
        ds_db3_libs_LIBS="$ds_db3_libs_LIBS -ldb3"
    ],
    [ ds_db3_libs_success=no ]
    )
AC_MSG_RESULT($ds_db3_libs_success)

if test "x$ds_db3_libs_success" != xyes
then
    AC_MSG_CHECKING([for db_create in -ldb3 (with -lpthread)])
    LIBS="$ds_db3_libs_LIBS -ldb3 -lpthread $LIBS"
    AC_LINK_IFELSE([AC_LANG_PROGRAM([[
#include <db.h>
        ]],
        [[
            DB** dummy_dbp = NULL;
            db_create(dummy_dbp, NULL, 0);
        ]])],
        [
            ds_db3_libs_success=yes
            ds_db3_libs_LIBS="$ds_db3_libs_LIBS -ldb3 -lpthread"
        ],
        [ ds_db3_libs_success=no ],
        )
    AC_MSG_RESULT($ds_db3_libs_success)
fi

if test x"$ds_db3_libs_success" = xyes
then
    AC_MSG_CHECKING([whether header and library versions are both db3])
    AC_RUN_IFELSE([AC_LANG_PROGRAM([[
#include <stdio.h>
#include <db.h>
        ]],
        [[
            int major, minor, patchlevel;
            int is_match;
            db_version(&major, &minor, &patchlevel);
            fprintf(stderr, "db version from header: %d.%d.%d\n",
                    DB_VERSION_MAJOR,
                    DB_VERSION_MINOR,
                    DB_VERSION_PATCH
                    );
            fprintf(stderr, "db version from library: %d.%d.%d\n",
                    major,
                    minor,
                    patchlevel
                    );
            if (major == DB_VERSION_MAJOR
                && minor == DB_VERSION_MINOR
                && patchlevel == DB_VERSION_PATCH
                )
            {
                is_match = 1;
            }
            else
            {
                is_match = 0;
            }
            return is_match ? 0 : 1;
        ]])],
        [ ds_db3_libs_success=yes],
        [ ds_db3_libs_success=no ]
        )
    AC_MSG_RESULT($ds_db3_libs_success)
fi

LIBS="$ds_db3_libs_save_LIBS"
if test x"$ds_db3_libs_success" = xyes
then
    ifelse([$1], [], [:], [$1="$ds_db3_libs_LIBS"])
    ifelse([$2], [], [:], [$2])
else
    ifelse([$3], [], [:], [$3])
fi
])

#
#   DS_BERKELEY_DB3([db_cppflags_out], [db_libs_out],
#                  [db_version], [db_version_major],
#                  [additional-action-if-found],
#                  [additional-action-if-not-found]
#                 )
AC_DEFUN(DS_BERKELEY_DB3,
[
ds_db3_save_CPPFLAGS="$CPPFLAGS"
ds_db3_save_LIBS="$LIBS"

ds_db3_CPPFLAGS=''
ds_db3_LIBS=''

ds_db3_success=yes

DS_BERKELEY_DB3_HEADERS([ds_db3_CPPFLAGS],
    [],
    [ds_db3_success=no])

CPPFLAGS="$ds_db3_CPPFLAGS $CPPFLAGS"
DS_BERKELEY_DB3_LIBS([ds_db3_LIBS],
    [],
    [ds_db3_success=no])

CPPFLAGS="$ds_db3_save_CPPFLAGS"
LIBS="$ds_db3_save_LIBS"

if test x"$ds_db3_success" = xyes
then
    ifelse([$1], [], [:], [$1="$ds_db3_CPPFLAGS"])
    ifelse([$2], [], [:], [$2="$ds_db3_LIBS"])

    # Forcing db_version_major to `3' is hack used untill proper db
    # version detection will implemented.
    ifelse([$4], [], [:], [$4=3])
    ifelse([$5], [], [:], [$5])
else
    ifelse([$6], [], [:], [$6])
fi
])
