/*
 DSPAM
 COPYRIGHT (C) 2003 NETWORK DWEEBS CORPORATION

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifdef HAVE_CONFIG_H
#include <auto-config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <dirent.h>
#include <db.h>

#ifdef TIME_WITH_SYS_TIME
#   include <sys/time.h>
#   include <time.h>
#else
#   ifdef HAVE_SYS_TIME_H
#       include <sys/time.h>
#   else
#       include <time.h>
#   endif
#endif

#if DB_VERSION_MINOR == 0
#       define DBOPEN(A, B, C, D, E, F, G) open(A, C, D, E, F, G)
#else
#       define DBOPEN(A, B, C, D, E, F, G) open(A, B, C, D, E, F, G)
#endif

#include "../libdb4_drv.h"
#include "../libdspam.h"
#include "../libdspam_objects.h"
#include "../config.h"
#include "../error.h"
#include "../storage_driver.h"
#include "../util.h"
#include "../language.h"

time_t start_time;
int process_directory(const char *directory);

int main(int argc, char *argv[]) {
  int rc;

  umask(006);
  start_time = time(NULL);

#ifdef DEBUG
  printf("purge start.\n");
#endif

  rc = process_directory(USERDIR);
  return rc < 0 ? EXIT_FAILURE : EXIT_SUCCESS;
}

/**
 *  @return
 *      0 on success or -1 on failure.
 */
int process_directory(const char *directory) {
  char filename[MAX_FILENAME_LENGTH];
  char scratch[MAX_FILENAME_LENGTH*3];
  char file[MAX_FILENAME_LENGTH];
  DBT key, data;
  DIR *dir;
  DB *dict, *newdict;
  DBC *c;
  int ret;
  struct dirent *entry;
  struct _libdb4_drv_spam_record record;
  struct stat s;
  int rc = 0;

  dir = opendir(directory);
  if (dir == NULL) {
    file_error(ERROR_FILE_OPEN, directory, strerror(errno));
    return EFILE;
  }

  while((entry = readdir(dir)) != NULL) {
    char *ext;

    if (!strcmp(entry->d_name, ".") || !strcmp(entry->d_name, "..")) 
      continue;

    snprintf(file, MAX_FILENAME_LENGTH, "%s/%s", directory, entry->d_name);
    if (stat(file, &s)) 
      continue;

    if (s.st_mode & S_IFDIR) {
      int subrc;
      subrc = process_directory(file);
      if (subrc < 0)
          rc = EFAILURE;
      continue;
    }

    if (strlen(entry->d_name)<=4)
      continue;

    ext = entry->d_name+(strlen(entry->d_name)-5);

    if (!strcmp(ext, ".dict")) {
      DSPAM_CTX *CTX;
      char *user;
      char *x;

      user = strdup(entry->d_name);
      x = strstr(user, ".dict");
      x[0] = 0;

      CTX = dspam_init(user, NULL, DSM_TOOLS, 0);

#ifdef DEBUG
      fprintf(stderr, "user identified as %s\n", user);
#endif
      if (_libdb4_drv_context_lock(CTX)) {
        fprintf(stderr, "unable to acquire context lock on user '%s': %s\n", user, strerror(errno));
        rc = EFAILURE;
        continue;
      }

#ifdef DEBUG
      fprintf(stderr, "opening %s/%s\n", USERDIR, entry->d_name);
#endif

      strcpy(filename, file);
      db_create(&dict, NULL, 0);

      ret = dict->set_cachesize(dict, 0, 1024*1024*DB_CACHESIZE_MB, 1);
      if (ret) {
        file_error(ERROR_FILE_WRITE, filename, db_strerror(ret));
        rc = EFAILURE;
        continue;
      }

      ret = dict->DBOPEN(dict, NULL, filename, NULL, DB_BTREE, 0, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
      if (ret) {
        file_error(ERROR_FILE_WRITE, filename, db_strerror(ret));
        _libdb4_drv_context_unlock(CTX);
        rc = EFAILURE;
        continue;
      }

      snprintf(filename, MAX_FILENAME_LENGTH, "%s.new", file);
      db_create(&newdict, NULL, 0);

      ret = newdict->DBOPEN(newdict, NULL, filename, NULL, DB_BTREE, DB_CREATE,  S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
      if (ret) {
        file_error(ERROR_FILE_WRITE, filename, db_strerror(ret));
        _libdb4_drv_context_unlock(CTX);
        rc = EFAILURE;
        continue;
      }

      ret = dict->cursor(dict, NULL, &c, 0);
      if (ret) {
        file_error(ERROR_CURSOR_CREATE, filename, db_strerror(ret));
        rc = EFAILURE;
        _libdb4_drv_context_unlock(CTX);
        (dict)->close(dict, 0);
        (newdict)->close(newdict, 0);
        unlink(filename);
        continue;
      }

      memset(&key, 0, sizeof(DBT));
      memset(&data, 0, sizeof(DBT));

      while( c->c_get(c, &key, &data, DB_NEXT)==0) {
        int delete = 0;

        if (!strncmp(key.data, "_TOTALS", 7)) {
#ifdef DEBUG
          struct _ds_spam_totals totals;
          memcpy(&totals, data.data, sizeof(struct _ds_spam_totals));
          fprintf(stderr, "Writing Totals: TS: %ld TI: %ld SM: %ld FP: %ld\n", totals.total_spam, totals.total_innocent, totals.spam_misses, totals.false_positives);
#endif

          ret = (*newdict->put)(newdict, NULL, &key, &data, 0);
          if (ret) { 
            fprintf(stderr, "put failed (at _TOTALS point)\n");
            rc = EFAILURE;
          }
          continue;
        } else 
          memcpy(&record, data.data, sizeof(struct _libdb4_drv_spam_record));
      
        if ((record.spam_hits + (record.innocent_hits*2))<5) {
          long delta = start_time-record.last_hit;
          if (delta>PURGE_BELOW_QUOTA) 
            delete = 1;
          if (record.spam_hits == 0 && delta>PURGE_NO_SPAM_HITS)
            delete = 1;
          if (record.spam_hits == 0 && record.innocent_hits == 1 && delta>PURGE_ONE_INNOCENT_HIT)
            delete = 1;
        }
    
        if (!delete) {
          ret = (*newdict->put)(newdict, NULL, &key, &data, 0);
          if (ret) { 
            fprintf(stderr, "put failed\n");
            rc = EFAILURE;
          }
        } else {
          delete = 0;
        }
      }
#ifdef DEBUG
      fprintf(stderr, "closing cursor\n");
#endif
      c->c_close(c);
#ifdef DEBUG
      fprintf(stderr, "closing databases\n");
#endif
      (dict)->close(dict, 0);
      (newdict)->close(newdict, 0);
      snprintf(scratch, sizeof(scratch), "%s %s.new %s", COPY, file, file);
      system(scratch);
      unlink(filename);

#ifdef DBEUG
      fprintf(stderr, "freeing lock\n");
#endif
      _libdb4_drv_context_unlock(CTX);
    }
  }
  closedir(dir);

#ifdef DEBUG
  fprintf(stderr, "complete.");
#endif

  return rc;
}
