/*
 DSPAM
 COPYRIGHT (C) 2003 NETWORK DWEEBS CORPORATION

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifdef HAVE_CONFIG_H
#include <auto-config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <dirent.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <db.h>
#include <errno.h>

#include "../libdspam.h"
#include "../libdspam_objects.h"
#include "../storage_driver.h"
#include "../config.h"
#include "../language.h"
#include "../util.h"
#include "../error.h"

#define TSYNTAX	"syntax: dspam_stats [username]"

int stat_user(const char *username);
int process_all_users(void);

int main(int argc, char **argv) {

  if (argc>1) {
    stat_user(argv[1]);
  } else {
    process_all_users();
  }

  exit(EXIT_SUCCESS);
}

int process_all_users(void) {
  DSPAM_CTX *CTX;
  char *user;

  CTX = dspam_init(NULL, NULL, DSM_TOOLS, 0);
  if (CTX == NULL) {
    fprintf(stderr, "Could not initialize context: %s\n", strerror(errno));
    exit(EXIT_FAILURE);
  }

  user = _ds_get_nextuser(CTX);
  while(user != NULL) {
    stat_user(user);
    user = _ds_get_nextuser(CTX);
  }

  dspam_destroy(CTX);
  return 0;
}

int stat_user(const char *username) {
  DSPAM_CTX *CTX;

  CTX = dspam_init(username, NULL, DSM_TOOLS, DSF_CLASSIFY);
  if (CTX == NULL) {
    fprintf(stderr, "Could not init context: %s\n", strerror(errno));
    return EUNKNOWN;
  }

  printf("%-16s  TS: %ld TI: %ld TM: %ld FP: %ld\n", username, CTX->totals.total_spam, CTX->totals.total_innocent, CTX->totals.spam_misses, CTX->totals.false_positives);

  dspam_destroy(CTX);
  return 0;
}

